/*******************/
/* System includes */
/*******************/

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <limits.h>

/******************/
/* EPICS includes */
/******************/

/* EPICS includes */
#include <drvIpac.h>
#include <iocsh.h>
#include <epicsExport.h>
#include <epicsThread.h>
#include <epicsString.h>
#include <epicsTimer.h>
#include <epicsExit.h>
#include <epicsMutex.h>
#include <epicsMessageQueue.h>
#include <cantProceed.h>
#include <devLib.h>
#include <errlog.h>
#include <asynPortDriver.h>

#ifndef DRVHY8414IPASYN_H
#define DRVHY8414IPASYN_H
#endif

/**************************************************************/
/* A couple of key parameters defining the ADC memory layout. */
#define BANKSIZE  				0x8000        	/* half back size in WORDs per channel: 32K */
#define MAX_MESSAGES  			100           	/* maximum number of messages */
#define MAX_ADC_CHANNELS    	16     			/* number of ADC channels on 8414 card */
#define HALF_MEM				0x80000			/* 512K Word*/
#define MAX_CARDS          		256          /* maxiumum number of cards in a system */

/* Identifiers in PROM. */
#define VITA4       			"VITA 4 "
#define HYTECID_LO  			0x0300
#define HYTECID_HI  			0x0080
#define PROM_MODEL  			0x8414

/* define bits in the 8414 CSR */
#define CSR_HF      0x0001          /* Set when lower conversion memory filled. */
#define CSR_F       0x0002          /* Set when upper conversion memory filled. */     
#define CSR_CC      0x0004          /* Set when requested sample count captured. */
#define CSR_MIS     0x0008          /* Set when memory update is inhibited. */
#define CSR_MII     0x0010          /* Set if interrupt generated by inhibit (MIS). */
#define CSR_EII     0x0020          /* Set to enable MIS interrupt. */
#define CSR_DA      0x0040          /* Set to allow the unit to disarm on completion of memory acquisition */
#define CSR_1M      0x0080          /* 1 => 64K samples, 0 => 128K samples per chan. */
#define CSR_HE      0x0100          /* Enable interrupt on HF set. */
#define CSR_FE      0x0200          /* Enable interrupt on F set. */
#define CSR_EE      0x0400          /* Enable end of sample sequence interrupt. */
#define CSR_ET      0x0800          /* Allow external enable to control ADC. */
#define CSR_XC      0x1000          /* 0 => use internal clock, 1 => external clock. */
#define CSR_ST      0x2000          /* Set to trigger capture of programmed samples. */
#define CSR_EX      0x4000          /* 0 => continuous capture, 1 => on trigger. */
#define CSR_ARM     0x8000          /* Set to enable capture into waveform RAM. */

#define EXTCSR_TEN  0x0001          /* Triggered status */
#define EXTCSR_SMI  0x0002          /* Software inhibit control and status. Write 1 to inhibit memory writing; write 0 to allow. Read 1 means hardware inhibit; read 0 means hardware allows */
#define EXTCSR_IS   0x0004          /* Single conversion completed flag */
#define EXTCSR_ISE  0x0008          /* Single conversion completed interrupt enable */
#define EXTCSR_CONT 0x0010          /* Allow multi trigger in CC mode */
#define EXTCSR_R  	0x0040          /* ADC range 0 = +/-10V and 1 = +/-5V */
#define EXTCSR_TE  	0x1000          /* Set 1 = external trigger on falling edge, 0 = raising edge */
#define EXTCSR_BTE 	0x2000          /* Set 1 = external trigger on both edges  */
#define EXTCSR_CAL	0x4000          /* Set 1 = no onboard calibration, 0 = yes, use onboard cali */

#define ENABLEADC 				1
#define DISABLEADC 				0

#define INIT 					0
#define RUNTIME 				1

/* for raw <=> voltage conversion */
#define NEG_FULL_VALUE 	0x000F
#define POSI_FULL_VALUE 0xFFF1
#define ZERO_VALUE		0x8000

/* driver support flags */
/* Bit0 - continuos mode support */
/* Bit1 - trigger mode support */
#define DRIVER_SUPPORT     		0x0003   		/* support continuous and trigger modes */

#define WAITING_TRIGGER         0           	/* waiting for trigger */
#define TRIGGERRED_COLLECTING   1           	/* triggerred and collecting data */
#define FINISHED_AVAILABLE      2           	/* finished and data available */

#ifndef USHORT 
#define USHORT unsigned short
#endif
#ifndef ULONG
#define ULONG unsigned int
#endif

/*****************************************************************************/
/*                                                                           */
/*                              Data Structures                              */
/*                                                                           */
/*****************************************************************************/


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* ADC Register and P#define hy8414SupportString			"SUPPORT"
ROM Layout                                              */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* The I/O memory is divided into two areas: register and PROM. We separate
 * them out for coding convenience. 
 */

/* This data structure maps the I/O registers */
typedef struct
{
    USHORT status;                  /* 0x00 */
    USHORT pointer;                 /* 0x02 */
    USHORT numConversions;          /* 0x04 */
    USHORT clockRate;               /* 0x06 */
    USHORT vector;                  /* 0x08 */
    USHORT extStatus;               /* 0x0A */
    USHORT gain;               		/* 0x0C */
    USHORT unused;               	/* 0x0E */
    USHORT data[MAX_ADC_CHANNELS];  /* 0x10 - 0x2E */
} REGISTERS;


/* This data structure maps the image of the ID PROM.  The PROM
 * starts at offset 0x80 from the base of the register memory. */
typedef struct
{
    char vita4[6];                  /* 0x80 */
    USHORT hytecIdHigh;             /* 0x86 */
    USHORT hytecIdLow;              /* 0x88 */
    USHORT model;                   /* 0x8A */
    USHORT revision;                /* 0x8C */
    USHORT unused1;                 /* 0x8E */
    USHORT driverIdHigh;            /* 0x90 */
    USHORT driverIdLow;             /* 0x92 */
    USHORT flags;                   /* 0x94 */
    USHORT byteCount;               /* 0x96 */
    USHORT calType;                 /* 0x98 */
    USHORT serialNumber;            /* 0x9A */
    USHORT unused2;                 /* 0x9C */
    USHORT WLO;                     /* 0x9E */
} ID_PROM;

/* Now put them together, the I/O memory map. */
typedef struct
{
    REGISTERS reg;                              /* 0x00 - 0x2F */
    char unused[0x80-sizeof(REGISTERS)];        /* 0x30 - 0x80 */
    ID_PROM prom;                               /* 0x80 - 0x9F */
} IO_MEMORY;

/* enumerated type of the driver's operational modes */
typedef enum{		
	hy8414Continuous,
	hy8414Triggered,
	hy8414Gated,
} hy8414Modes;


/*****************************************************************************/
/*                                                                           */
/*                      Interrupt/Queue Routines	                         */
/*                                                                           */
/*****************************************************************************/

extern "C" {
static void rebootCallbackC     ( void *drvPvt );
static void INTSRC              ( int cardNumber ); 	/* Interrupt Service Routine */
static void intQueuedTaskC      ( void *drvPvt );
static void intSweepQueuedTaskC	( void *drvPvt );
}

/** This is the class definition for the Hy8414Asyn class
  */
class Hy8414Asyn : public asynPortDriver {
public:
	Hy8414Asyn(char *portName, int carrierNum, int ipSlotNum, int vectorNum, int mode);

    /* These are the methods that we override from asynPortDriver */
    virtual asynStatus writeInt32(asynUser *pasynUser, epicsInt32 value);
    virtual asynStatus readInt32(asynUser *pasynUser, epicsInt32 *value);
    virtual asynStatus readInt32Array(asynUser *pasynUser, epicsInt32 *value,
                                        size_t nElements, size_t *nIn);
    virtual asynStatus getBounds(asynUser *pasynUser, epicsInt32 *low, epicsInt32 *high);
    virtual asynStatus readFloat64(asynUser *pasynUser, epicsFloat64 *value);
    virtual asynStatus readFloat64Array(asynUser *pasynUser, epicsFloat64 *value,
                                        size_t nElements, size_t *nIn);
    virtual void report(FILE *fp, int details);
    
    /* these are specific to Hy8414Asyn class */
    int Hy8414ExtInit(int sample, int average, int offset, int clockRate, 
    		int extClock, int fastADC, int gain, int range);	
    void rebootCallback();
	void intSR(); 				/* Interrupt Service Routine */
	void intQueuedTask();
	void intSweepQueuedTask();

protected:
    int hy8414SweepData_;
    #define FIRST_8414_PARAM hy8414SweepData_
    int hy8414FloatSweepData_;
	int hy8414Data_;
	int hy8414FloatData_;
	int hy8414ScanPeriod_;
	int hy8414SetAverage_;
	int hy8414SetSample_;
	int hy8414SetMode_;
	int hy8414SetOffset_;
	int hy8414Enable_;
	int hy8414SoftTrigger_;
	int hy8414AutoReenable_;
	int hy8414ClearBuffer_;
	int hy8414SetClockRate_;
	int hy8414SetExtClock_;
	int hy8414Csr_;
	int hy8414ExtCsr_;

	int hy8414Gain_;
	int hy8414Range_;
	int hy8414FastADC_;

	int hy8414AverageOverflow_;
	int hy8414Overflow_;
	int hy8414BufferCount_;
	int hy8414TriggerStatus_;
	int hy8414FWVersion_;
	int hy8414DriverVersion_;
	int hy8414Support_;
    #define LAST_8414_PARAM hy8414Support_

private:
	void setupADC               ( int runtime );
	int setClockRate            ( int runtime );
	int setExtClock             ( int runtime );
	int softwareTrigger         ();
	int setSamples				( int runtime );
	int checkAverageOverflow 	( int runtime );
	int setOffset				( int runtime );
	int setMode           		( int runtime );
	int setGain           		( int runtime );
	int setRange           		( int runtime );
	int setEnableDisable		( int enable );
	double toVoltage            ( double rawval );
    
	volatile IO_MEMORY 	*adc;			   	/* pointer to IP card A16 space registers, ID and calibration data */ 
	volatile USHORT 	*convMemoryBase;   	/* pointer to IP card memory space */

    ULONG 	rawData[MAX_ADC_CHANNELS][BANKSIZE*2];/* ADC data for I/O Intr Int32Array */
    double 	adcData[MAX_ADC_CHANNELS][BANKSIZE*2];/* ADC data for I/O Intr Float64Array */
    ULONG  	chanData[MAX_ADC_CHANNELS];    	/* storage for the IP card last channel data values */

	/* Storage for the current Configuration parameters */

    USHORT	carrierNum;		                /* carrier index number when adding carrier */
    USHORT	ipSlotNum;		                /* IP card slot number on IP carrier - boot time configuration only */
    USHORT	vectorNum;		                /* Interrupt vector number - boot time configuration only */

	int		enable;							/* for ISR */
	int		mode;							/* for ISR */
	int		samples;						/* for ISR */
	int		fastadc;						/* for ISR */
	
	int		banksize;
	int 	lastPointer;
    int     sampleCounter;                  /* counter for continuous mode to count samples acquired in order to generate callbacks */

/* driver specific parameters */
    int		rebooting;		                /* Flag to indicate is the IOC is currently re-booting */

/* Work Queue Stuff */
    epicsMessageQueueId intMsgQId;
    int messagesSent;                       /* for report */
    int messagesFailed;                     /* for report */
    epicsMessageQueueId intSweepQId;
};

#define hy8414SweepDataString	    "DATA"
#define hy8414FloatSweepDataString	"FLOATDATA"
#define hy8414DataString	        "GENDATA"
#define hy8414FloatDataString	    "FLOATGENDATA"
#define hy8414ScanPeriodString	    "SCAN_PERIOD"
#define hy8414SetAverageString	    "SETAVERAGE"
#define hy8414SetSampleString	    "SETSAMPLE"
#define hy8414SetModeString		    "SETMODE"
#define hy8414SetOffsetString	    "SETOFFSET"
#define hy8414EnableString		    "ENABLE"
#define hy8414SoftTriggerString		"SOFTTRIGGER"
#define hy8414AutoReenableString	"REENABLE"
#define hy8414ClearBufferString		"CLEARBUFFER"
#define hy8414SetClockRateString	"SETCLOCKRATE"
#define hy8414SetExtClockString		"SETEXTCLOCK"
#define hy8414CsrString				"CSR"
#define hy8414ExtCsrString			"EXTCSR"

#define hy8414GainString			"GAIN"
#define hy8414RangeString			"RANGE"
#define hy8414FastADCString			"FASTADC"

#define hy8414AverageOverflowString	"AVERAGEOVERFLOW"
#define hy8414OverflowString		"OVERFLOW"
#define hy8414BufferCountString		"BUFFERCOUNT"
#define hy8414TriggerStatusString	"TRIGGERSTATE"
#define hy8414FWVersionString		"FWVERSION"
#define hy8414DriverVersionString	"DRIVERVERSION"
#define hy8414SupportString			"SUPPORT"

#define NUM_PARAMS (&LAST_8414_PARAM - &FIRST_8414_PARAM + 1)

/** Configuration command, called directly or from iocsh */
/* 
 * Where: -
 *     portName = asyn port name
 *   carrierNum = carrier index number when adding a carrier card by ipacAddHy8002 call
 *    ipSlotNum = IP slot number on the carrier board (0 = A, 1 = B, 2 = C, 3 = D, 4 = E, 5 = F),
 *    vectorNum = interrupt vector (0 = find one for me),
 *         mode = 0 = continuous, 1 = triggered, 3 = gated,
 */
extern "C" int Hy8414AsynInit(char *portName, int vmeSlotNum, int ipSlotNum, int vectorNum, 
					int mode)
{
    Hy8414Asyn *pHy8414Asyn = new Hy8414Asyn(portName, vmeSlotNum, ipSlotNum, vectorNum, mode);
    pHy8414Asyn = NULL;  /* This is just to avoid compiler warnings */
    return(asynSuccess);
}

/** Configuration command, called directly or from iocsh */
/* 
 * Where: -
 *     portName = asyn port name
 *      samples = number of samples (not applicable in gated modes),
 *		average = number of readings to SUM for average
 *       offset = Trigger mode: Offset from trigger point to first ADC reading to return
 *				  Gated mode:   Offset from the start (or end if negative) of the gated period to the ADC readings to return
 *				  Continuous:   Not applicable
  *    clockRate = 0  = 1Hz,    1 = 2Hz,    2 = 5Hz,    3 = 10Hz,
 *				  4  = 20Hz,   5 = 50Hz,   6 = 100Hz,  7 = 200Hz,
 *				  8  = 500Hz,  9 = 1kHz,  10 = 2kHz,  11 = 5kHz,
 *				  12 = 10khz, 13 = 20Khz, 14 = 50kHz, 15 = 100kHz.
 *     extClock = 0 = internal, 1 = extrenal,
 *		fastAdc = flag to say if the ADC is running at fast speed for normal use. =1 fast, =0 slow for MCA & EPID records 
 *		   gain = 0/1/2/3  for 1/2/4/8 times
 *		  range = 0 : +/-10V, 1 : +/-5V
 */
extern "C" int Hy8414AsynExtInit(char *portName, int sample, int average, int offset, 
		int clockRate, int extClock, int fastADC, int gain, int range)
{
    asynStatus status;
    asynInterface *pasynInterface;
    asynUser *pasynUser;
    Hy8414Asyn *pHy8414 = NULL;

    pasynUser = pasynManager->createAsynUser(0, 0);
    status = pasynManager->connectDevice(pasynUser, portName, 0);
    if (status != asynSuccess) {
        errlogPrintf("Hy8414AsynExtInit, error in connectDevice %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pasynInterface = pasynManager->findInterface(pasynUser, asynFloat64Type, 1);
    if (!pasynInterface) {
        errlogPrintf("Hy8414AsynExtInit cannot find Float64 interface %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pHy8414 = (Hy8414Asyn*) pasynInterface->drvPvt;
    pHy8414->Hy8414ExtInit(sample, average, offset, clockRate, extClock, fastADC, gain,	range);
    return(asynSuccess);
} 
				
// Init
static const iocshArg initArg0 = { "Port name",iocshArgString};
static const iocshArg initArg1 = { "vmeSlotNum",iocshArgInt};
static const iocshArg initArg2 = { "ipSlotNum",iocshArgInt};
static const iocshArg initArg3 = { "vectorNum",iocshArgInt};
static const iocshArg initArg4 = { "Scan mode", iocshArgInt};
static const iocshArg * const initArgs[] = {&initArg0,
                                             &initArg1,
                                             &initArg2,
                                             &initArg3,
                                             &initArg4};
static const iocshFuncDef initFuncDef = {"Hy8414AsynInit",5,initArgs};
static void initCallFunc(const iocshArgBuf *args)
{
    Hy8414AsynInit(args[0].sval, args[1].ival, args[2].ival, args[3].ival,
					args[4].ival);
}

// Extension init
static const iocshArg initArg0Ext = { "Port name",iocshArgString};
static const iocshArg initArg1Ext = { "samples",iocshArgInt};
static const iocshArg initArg2Ext = { "average",iocshArgInt};
static const iocshArg initArg3Ext = { "offset",iocshArgInt};
static const iocshArg initArg4Ext = { "clockRate",iocshArgInt};
static const iocshArg initArg5Ext = { "External Clock", iocshArgInt};
static const iocshArg initArg6Ext = { "fastADC", iocshArgInt};
static const iocshArg initArg7Ext = { "gain", iocshArgInt};
static const iocshArg initArg8Ext = { "Range",iocshArgInt};
static const iocshArg * const initArgsExt[] = {&initArg0Ext,
                                             &initArg1Ext,
                                             &initArg2Ext,
                                             &initArg3Ext,
                                             &initArg4Ext,
                                             &initArg5Ext,
                                             &initArg6Ext,
                                             &initArg7Ext,
                                             &initArg8Ext};
static const iocshFuncDef initFuncDefExt = {"Hy8414AsynExtInit",9,initArgsExt};
static void initCallFuncExt(const iocshArgBuf *args)
{
    Hy8414AsynExtInit(args[0].sval, args[1].ival, args[2].ival, args[3].ival,
					args[4].ival, args[5].ival, args[6].ival, args[7].ival,
					args[8].ival);
}

void drvHy8414AsynRegister(void)
{
    iocshRegister(&initFuncDef,initCallFunc);
    iocshRegister(&initFuncDefExt,initCallFuncExt);
}

extern "C" {
epicsExportRegistrar(drvHy8414AsynRegister);
}

